import zipfile
import os
import tarfile
import subprocess

from memorious.util import random_filename

ZIP_MIME_TYPES = [
    'application/zip',
    'application/x-zip',
    'multipart/x-zip',
    'application/zip-compressed',
    'application/x-zip-compressed',
]

TAR_MIME_TYPES = [
    'application/tar',
    'application/x-tar',
    'application/x-tgz',
    'application/x-gtar',
    'application/x-gzip',
]

SEVENZIP_MIME_TYPES = [
    'application/x-7z-compressed',
    'application/7z-compressed'
]


def extract_zip(file_path, extract_dir):
    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(extract_dir)
        extracted_files = []
        for name in zip_ref.namelist():
            file_path = os.path.join(extract_dir, name)
            if os.path.isfile(file_path):
                extracted_files.append(file_path)
        return extracted_files


def extract_tar(file_path, extract_dir, context):
    with tarfile.open(file_path, "r:*") as tar_ref:
        extracted_files = []
        for name in tar_ref.getnames():
            # Make it safe. See warning at
            # https://docs.python.org/2/library/tarfile.html#tarfile.TarFile.extractall  # noqa
            if name.startswith("..") or name.startswith("/"):
                context.log.info(
                    "Bad path %s while extracting archive at %s",
                    name, file_path
                )
            else:
                tar_ref.extract(name, extract_dir)
                file_path = os.path.join(extract_dir, name)
                if os.path.isfile(file_path):
                    extracted_files.append(file_path)
        return extracted_files


def extract_7zip(file_path, extract_dir, context):
    return_code = subprocess.call(
        ['7z', 'x', file_path, '-r', '-o%s' % extract_dir])
    if return_code != 0:
        context.log.warning(
            "Couldn't extract file: %s", file_path
        )
        return
    extracted_files = []
    for root, directories, filenames in os.walk(extract_dir):
        for filename in filenames:
            extracted_files.append(os.path.join(root, filename))
    return extracted_files


def extract(context, data):
    """Extract a compressed file"""
    with context.http.rehash(data) as result:
        file_path = result.file_path
        content_type = result.content_type
        extract_dir = random_filename(context.work_path)
        if content_type in ZIP_MIME_TYPES:
            extracted_files = extract_zip(file_path, extract_dir)
        elif content_type in TAR_MIME_TYPES:
            extracted_files = extract_tar(file_path, extract_dir, context)
        elif content_type in SEVENZIP_MIME_TYPES:
            extracted_files = extract_7zip(file_path, extract_dir, context)
        else:
            context.log.warning(
                "Unsupported archive content type: %s", content_type
            )
            return
        extracted_content_hashes = {}
        for path in extracted_files:
            relative_path = os.path.relpath(path, extract_dir)
            content_hash = context.store_file(path)
            extracted_content_hashes[relative_path] = content_hash
            data['content_hash'] = content_hash
            data['file_name'] = relative_path
            context.emit(data=data.copy())


import json

from sqlalchemy.types import TypeDecorator, TEXT
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import JSONB

Base = declarative_base()


class TextJSON(TypeDecorator):
    """Stores and retrieves JSON as TEXT."""
    impl = TEXT

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        return json.loads(value)


JSON = TextJSON().with_variant(JSONB, 'postgresql')



import logging
from datetime import datetime
from sqlalchemy import Column, String, Integer, DateTime, ForeignKey

from memorious.core import session
from memorious.model.common import Base

log = logging.getLogger(__name__)


class Event(Base):
    """Document the start time, end time, type and outcome of a given task."""
    __tablename__ = 'event'

    LEVEL_WARNING = 'warning'
    LEVEL_ERROR = 'error'

    id = Column(Integer, primary_key=True)
    level = Column(String, nullable=False, index=True)
    crawler = Column(String, nullable=False, index=True)
    stage = Column(String, nullable=True, index=True)
    run_id = Column(String, nullable=True, index=True)
    error_type = Column(String, nullable=True)
    error_message = Column(String, nullable=True)
    error_details = Column(String, nullable=True)
    timestamp = Column(DateTime, default=datetime.utcnow)

    @classmethod
    def save(cls, crawler, stage, level, run_id, error_type=None,
             error_message=None, error_details=None):
        """Create an event, possibly based on an exception."""
        event = cls()
        event.crawler = crawler
        event.stage = stage
        event.level = level
        event.run_id = run_id
        event.error_type = error_type
        event.error_message = error_message
        event.error_details = error_details
        session.add(event)
        return event

    @classmethod
    def delete(cls, crawler):
        pq = session.query(cls)
        pq = pq.filter(cls.crawler == crawler)
        pq.delete(synchronize_session=False)

    def __repr__(self):
        return '<Event(%s,%s,%s,%s)>' % \
            (self.crawler, self.stage, self.error_type, self.level)



import logging
from datetime import datetime
from sqlalchemy import Column, String, Integer, DateTime

from memorious.core import session
from memorious.model.common import Base, JSON

log = logging.getLogger(__name__)


class Tag(Base):
    """A simple key/value table used to store interim results."""
    __tablename__ = 'tag'

    id = Column(Integer, primary_key=True)
    crawler = Column(String(255), nullable=False, index=True)
    key = Column(String, nullable=False, index=True)
    value = Column(JSON, nullable=False, default={})
    timestamp = Column(DateTime, default=datetime.utcnow)

    @classmethod
    def save(cls, crawler, key, value):
        obj = cls.find(crawler, key)
        if obj is None:
            obj = cls()
            obj.crawler = crawler.name
            obj.key = key
        obj.value = value
        obj.timestamp = datetime.utcnow()
        session.add(obj)
        return obj

    @classmethod
    def find(cls, crawler, key, since=None):
        q = session.query(cls)
        q = q.filter(cls.crawler == crawler.name)
        q = q.filter(cls.key == key)
        if since is not None:
            q = q.filter(cls.timestamp >= since)
        q = q.order_by(cls.timestamp.desc())
        return q.first()

    @classmethod
    def exists(cls, crawler, key, since=None):
        q = session.query(cls)
        q = q.filter(cls.crawler == crawler.name)
        q = q.filter(cls.key == key)
        if since is not None:
            q = q.filter(cls.timestamp >= since)
        return q.count() > 0

    @classmethod
    def delete(cls, crawler):
        pq = session.query(cls)
        pq = pq.filter(cls.crawler == crawler)
        pq.delete(synchronize_session=False)

    def __repr__(self):
        return '<Tag(%s,%s)>' % (self.crawler, self.key)



import tempfile
import os
import shutil
from datetime import datetime
import mimetypes
import logging

from datafreeze.app import freeze
import boto3

from memorious.core import datastore
from memorious.settings import ARCHIVE_AWS_KEY_ID, ARCHIVE_AWS_SECRET


log = logging.getLogger(__name__)
TAG_LATEST = 'latest'
PREFIX = 'data'


def s3_upload(bucket, file_path):
    """
    Upload a given file to a s3 bucket. Stores 2 versions: 1 timestamped
    version and one latest version.
    """
    session = boto3.Session(aws_access_key_id=ARCHIVE_AWS_KEY_ID,
                            aws_secret_access_key=ARCHIVE_AWS_SECRET)
    client = session.client('s3')
    file_name = os.path.basename(file_path)
    mime_type, _ = mimetypes.guess_type(file_name)
    mime_type = mime_type or 'application/octet-stream'
    tag = datetime.utcnow().date().isoformat()
    key_name = os.path.join(PREFIX, tag, file_name)
    log.info("Uploading [%s]: %s", bucket, key_name)
    args = {
        'ContentType': mime_type,
        'ACL': 'public-read',
    }
    client.upload_file(file_path, bucket, key_name, ExtraArgs=args)
    copy_name = os.path.join(PREFIX, TAG_LATEST, file_name)
    copy_source = {'Key': key_name, 'Bucket': bucket}
    client.copy(copy_source, bucket, copy_name, ExtraArgs=args)
    log.info('File available at http://%s/%s', bucket, key_name)


def export_tables(params):
    """Dump tables to csv files and upload to s3 bucket"""
    tempdir = tempfile.mkdtemp()
    for entry in params:
        table = datastore[entry["table"]]
        filename = entry.get("csv_filename", entry["table"] + ".csv")
        bucket = entry.get("bucket")
        freeze(table.all(), filename=filename, prefix=tempdir)
        file_path = os.path.join(tempdir, filename)
        s3_upload(bucket, file_path)
    shutil.rmtree(tempdir)



import os
import io
import yaml
import logging
import time
from datetime import timedelta, datetime
from importlib import import_module

from memorious import settings, signals
from memorious.core import session, local_queue
from memorious.model import Tag, Event, Result
from memorious.reporting import get_last_run, is_running, cleanup_crawler
from memorious.logic.context import handle
from memorious.logic.stage import CrawlerStage

log = logging.getLogger(__name__)


class Crawler(object):
    """A processing graph that constitutes a crawler."""
    SCHEDULES = {
        'daily': timedelta(days=1),
        'weekly': timedelta(weeks=1),
        'monthly': timedelta(weeks=4)
    }

    def __init__(self, manager, source_file):
        self.manager = manager
        self.source_file = source_file
        with io.open(source_file, encoding='utf-8') as fh:
            self.config_yaml = fh.read()
            self.config = yaml.load(self.config_yaml)

        self.name = os.path.basename(source_file)
        self.name = self.config.get('name', self.name)
        self.description = self.config.get('description', self.name)
        self.category = self.config.get('category', 'scrape')
        self.schedule = self.config.get('schedule')
        self.disabled = self.config.get('disabled', False)
        self.init_stage = self.config.get('init', 'init')
        self.delta = Crawler.SCHEDULES.get(self.schedule)
        self.delay = int(self.config.get('delay', 0))
        self.expire = int(self.config.get('expire', settings.EXPIRE))
        self.stealthy = self.config.get('stealthy', False)
        self.cleanup_config = self.config.get('cleanup', {})

        self.stages = {}
        for name, stage in self.config.get('pipeline', {}).items():
            self.stages[name] = CrawlerStage(self, name, stage)

    def check_due(self):
        """Check if the last execution of this crawler is older than
        the scheduled interval."""
        if self.disabled:
            return False
        if self.delta is None:
            return False
        last_run = get_last_run(self)
        if last_run is None:
            return True
        now = datetime.utcnow()
        if now > last_run + self.delta:
            return True
        return False

    def flush(self):
        """Delete all run-time data generated by this crawler."""
        Tag.delete(self.name)
        Event.delete(self.name)
        Result.delete(self.name)
        session.commit()
        signals.crawler_flush.send(self)

    def run(self, incremental=None):
        """Queue the execution of a particular crawler."""
        state = {
            'crawler': self.name,
            'incremental': settings.INCREMENTAL
        }
        if incremental is not None:
            state['incremental'] = incremental
        stage = self.get(self.init_stage)
        handle.delay(state, stage.name, {})

        # If running in eager mode, we need to block until all the queued
        # tasks are finished.
        while not local_queue.is_empty:
            time.sleep(1)

    def replay(self, stage):
        """Re-run all tasks issued to a particular stage.

        This sort of requires a degree of idempotence for each operation.
        Usually used to re-parse a set of crawled documents.
        """
        query = Result.by_crawler_next_stage(self.name, stage)
        for result in query:
            state = {'crawler': self.name}
            handle.delay(state, stage, result.data)

    @property
    def cleanup_method(self):
        if self.cleanup_config:
            method = self.cleanup_config["method"]
            package = 'memorious.helpers.export'
            module = import_module(package)
            return getattr(module, method)

    def cleanup(self):
        """Run a cleanup method after the crawler finishes running"""
        should_run_cleanup = False
        # Run cleanup if the crawler has finished running
        if not is_running(self):
            should_run_cleanup = True
        # Run cleanup if the last operation of the crawler was more than half
        # a day ago and it's just hanging in running state since then.
        delta = timedelta(hours=12)
        last_run = get_last_run(self)
        if last_run is not None:
            now = datetime.utcnow()
            if now > last_run + delta:
                should_run_cleanup = True

        if should_run_cleanup:
            cleanup_crawler(self)
            if self.cleanup_method:
                log.info("Running clean up for %s" % self.name)
                self.cleanup_method(self.cleanup_config["params"])
            else:
                pass

    def get(self, name):
        return self.stages.get(name)

    def __iter__(self):
        return iter(self.stages.values())

    def __repr__(self):
        return '<Crawler(%s)>' % self.name



import json
import requests
from pprint import pprint  # noqa
from banal import clean_dict
from six.moves.urllib.parse import urljoin

from memorious import settings


def aleph_emit(context, data):
    context.log.info("Store [aleph]: %s", data.get('url'))
    if not settings.ALEPH_HOST:
        context.log.warning("No $MEMORIOUS_ALEPH_HOST, skipping upload...")
        return
    if not settings.ALEPH_API_KEY:
        context.log.warning("No $MEMORIOUS_ALEPH_API_KEY, skipping upload...")
        return

    with context.http.rehash(data) as result:
        if not result.ok:
            return
        submit_result(context, result, data)


def submit_result(context, result, data):
    if result.file_path is None:
        context.log.info("Cannot ingest response: %s", result)
        return

    session = requests.Session()
    session.headers['Authorization'] = 'apikey %s' % settings.ALEPH_API_KEY
    collection_id = get_collection_id(context, session)
    if collection_id is None:
        return

    languages = context.params.get('languages', [])
    countries = context.params.get('countries', [])
    mime_type = context.params.get('mime_type', result.content_type)
    meta = {
        'crawler': context.crawler.name,
        'source_url': data.get('source_url', result.url),
        'title': data.get('title'),
        'author': data.get('author'),
        'file_name': data.get('file_name'),
        'foreign_id': data.get('foreign_id', result.request_id),
        'mime_type': data.get('mime_type', mime_type),
        'countries': data.get('countries', countries),
        'languages': data.get('languages', languages),
        'retrieved_at': data.get('retrieved_at', result.retrieved_at),
        'modified_at': data.get('modified_at', result.last_modified),
        'published_at': data.get('published_at'),
        'headers': dict(result.headers or {})
    }

    if data.get('parent_foreign_id'):
        meta['parent'] = {'foreign_id': data.get('parent_foreign_id')}
    if not data.get('file_name') and result.file_name:
        meta['file_name'] = result.file_name

    meta = clean_dict(meta)
    # pprint(meta)

    url = make_url('collections/%s/ingest' % collection_id)
    title = meta.get('title', meta.get('file_name', meta.get('source_url')))
    context.log.info("Sending '%s' to %s", title, url)
    file = open(result.file_path, 'rb')
    res = session.post(url,
                       data={'meta': json.dumps(meta)},
                       files={'file': file})
    if not res.ok:
        context.emit_warning("Could not ingest '%s': %r" % (title, res.text))
    else:
        document = res.json().get('documents')[0]
        context.log.info("Ingesting, document ID: %s", document['id'])


def get_collection_id(context, session):
    url = make_url('collections')
    if hasattr(context.stage, '_aleph_cid'):
        return context.stage._aleph_cid
    foreign_id = context.get('collection', context.crawler.name)
    res = session.get(url, params={
        'filter:foreign_id': foreign_id
    })
    data = res.json()
    for coll in data.get('results'):
        if coll.get('foreign_id') == foreign_id:
            context.stage._aleph_cid = coll.get('id')
            return context.stage._aleph_cid

    res = session.post(url, json={
        'label': context.crawler.description,
        'category': context.crawler.category,
        'managed': True,
        'foreign_id': foreign_id
    })

    if not res.ok:
        context.log.info("Could not create collection: %s", foreign_id)
        return

    context.stage._aleph_cid = res.json().get('id')
    return context.stage._aleph_cid


def make_url(path):
    prefix = urljoin(settings.ALEPH_HOST, '/api/2/')
    return urljoin(prefix, path)



import os
from uuid import uuid4
from normality import stringify


def make_key(*criteria):
    """Make a string key out of many criteria."""
    criteria = [stringify(c) for c in criteria]
    criteria = [c for c in criteria if c is not None]
    if len(criteria):
        return ':'.join(criteria)


def random_filename(path=None):
    """Make a UUID-based file name which is extremely unlikely
    to exist already."""
    filename = uuid4().hex
    if path is not None:
        filename = os.path.join(path, filename)
    return filename



from collections import namedtuple
import pytest
from memorious.helpers.rule import RULES, Rule, DomainRule, UrlPatternRule


spec = {
    'and': [
        {'domain': 'occrp.org'},
        {'not': {
            'or': [
                {'domain': 'vis.occrp.org'},
                {'domain': 'tech.occrp.org'},
                {'domain': 'data.occrp.org'},
                {'mime_group': 'assets'},
                {'mime_group': 'images'},
                {'pattern': 'https://www.occrp.org/en/component/.*'},
                {'pattern': 'https://www.occrp.org/en/donate.*'},
                {'pattern': 'https://www.occrp.org/.*start=.*'},
                {'pattern': 'https://www.occrp.org/ru/.*'}
            ]
        }}
    ]
}

invalid_spec = {
    "and": [
        {'domain': 'occrp.org'},
        {"not": {'domain': 'vis.occrp.org'}}
    ],
    "not": {'mime_group': 'images'},
}

rule = {}


class TestRule(object):
    def test_get_rule(self):
        with pytest.raises(Exception):
            Rule.get_rule(invalid_spec)
        assert isinstance(Rule.get_rule(spec), RULES["and"])


class TestDomainRule(object):
    def test_domain_rule(self):
        rule = DomainRule("occrp.org")
        rule.configure()
        Response = namedtuple("Response", "url")
        res = Response(url="http://occrp.org")
        assert rule.apply(res)
        res = Response(url="http://not-occrp.org")
        assert rule.apply(res) is False


class TestUrlPatternRule(object):
    def test_url_pattern(self):
        rule = UrlPatternRule("https://www.occrp.org/en/donate.*")
        rule.configure()
        Response = namedtuple("Response", "url")
        res = Response(url="https://www.occrp.org/en/donate.html")
        assert rule.apply(res)
        res = Response(url="http://not-occrp.org")
        assert rule.apply(res) is False



import pytz
import dateparser
from datetime import datetime


def parse_date(text, format_hint=None):
    if text is None:
        return
    if format_hint is not None:
        parsed = datetime.strptime(text, format_hint)
    else:
        # Strip things that don't belong in dates but websites like to wrap
        # their dates with them anyway
        cleaned = text.strip('[] ')
        parsed = dateparser.parse(cleaned)
    return naive_datetime(parsed)


def iso_date(text, format_hint=None):
    parsed = parse_date(text, format_hint)
    if parsed is not None:
        return parsed.isoformat()


def naive_datetime(dt):
    if not isinstance(dt, datetime):
        return dt
    if dt.tzinfo is not None:
        dt = dt.astimezone(pytz.utc)
        dt = dt.replace(tzinfo=None)
    return dt



import logging
import contextlib
import datetime

from redis_rate_limit import RateLimit, TooManyRequests

from memorious import settings
from memorious.core import redis_pool
from memorious.exc import RateLimitException

log = logging.getLogger(__name__)
global_call_log = {}


@contextlib.contextmanager
def rate_limiter(context):
    resource = "%s:%s" % (context.crawler.name, context.stage.name)
    rate_limit = context.stage.rate_limit
    if settings.REDIS_HOST:
        try:
            with RateLimit(resource=resource,
                           client='memorious',
                           max_requests=rate_limit,
                           redis_pool=redis_pool):
                yield
        except TooManyRequests:
            raise RateLimitException
    else:
        if resource in global_call_log:
            last_called = global_call_log[resource]
            diff = (datetime.datetime.now() - last_called).total_seconds()
            if diff < (1.0/rate_limit):
                raise RateLimitException
        global_call_log[resource] = datetime.datetime.now()
        yield



import json
import six
from memorious.logic.context import Context, handle
from memorious.model import Event
from memorious.core import local_queue


class TestContext(object):
    def test_context(self, context):
        assert isinstance(context.run_id, six.string_types)

    def test_content_hash(self, context):
        content_hash = context.store_data(json.dumps({"hello": "world"}))
        assert isinstance(content_hash, six.string_types)
        with context.load_file(content_hash) as fh:
            assert hasattr(fh, "read")

    def test_dump_load_state(self, context, crawler, stage):
        dump = context.dump_state()
        new_context = Context.from_state(dump, stage.name)
        assert isinstance(new_context, Context)
        assert new_context.run_id == context.run_id
        assert new_context.crawler.name == crawler.name
        assert new_context.stage.name == stage.name
        assert all(
            (k, v) in new_context.state.items()
            for k, v in context.state.items()
        )

    def test_emit_exception(self, context):
        exc = Exception("Uh oh!")
        event = context.emit_exception(exc)
        assert isinstance(event, Event)
        assert event.level == Event.LEVEL_ERROR

    def test_emit_warning(self, context):
        msg = "hello"
        event = context.emit_warning(msg)
        assert isinstance(event, Event)
        assert event.level == Event.LEVEL_WARNING

    def test_execute(self, context, mocker):
        data = {"answer": 42}
        mocker.patch("memorious.logic.stage.CrawlerStage.method",
                     new_callable=mocker.PropertyMock)
        context.execute(data)
        assert context.stage.method.call_count == 1
        context.stage.method.assert_called_once_with(context, data)


def test_handle_execute(stage, context, mocker):
    mocker.patch.object(Context, "from_state", return_value=context)
    mocker.patch.object(local_queue, "queue_operation")
    data = {"hello": "world"}
    handle({"foo": "bar"}, stage, data)
    assert local_queue.queue_operation.call_count == 1
    local_queue.queue_operation.assert_called_once_with(context, data)



import os
import six
import uuid
import shutil
import random
import logging
import traceback
from copy import deepcopy
from tempfile import mkdtemp
from datetime import datetime, timedelta
from contextlib import contextmanager
import time

from memorious.core import manager, storage, celery, session
from memorious.core import datastore, local_queue
from memorious.model import Result, Tag, Event
from memorious.exc import StorageFileMissing
from memorious.logic.http import ContextHttp
from memorious.logic.rate_limit import rate_limiter, RateLimitException
from memorious.util import make_key, random_filename
from memorious import settings, signals


class Context(object):
    """Provides state tracking and methods for operation interactions."""

    def __init__(self, crawler, stage, state):
        self.crawler = crawler
        self.stage = stage
        self.state = state
        self.params = stage.params
        self.incremental = state.get('incremental')
        self.run_id = state.get('run_id') or uuid.uuid1().hex
        self.work_path = mkdtemp()
        self.log = logging.getLogger('%s.%s' % (crawler.name, stage.name))
        self.http = ContextHttp(self)
        self.datastore = datastore

    def get(self, name, default=None):
        """Get a configuration value and expand environment variables."""
        value = self.params.get(name, default)
        if isinstance(value, six.string_types):
            value = os.path.expandvars(value)
        return value

    def emit(self, rule='pass', stage=None, data={}, delay=None):
        """Invoke the next stage, either based on a handling rule, or by calling
        the `pass` rule by default."""
        if stage is None:
            stage = self.stage.handlers.get(rule)
        if stage is None or stage not in self.crawler.stages:
            self.log.info("No next stage: %s (%s)" % (stage, rule))
            return
        state = self.dump_state()
        delay = delay or self.crawler.delay
        Result.save(self.crawler, self.stage.name, stage, data)
        handle.apply_async((state, stage, data), countdown=delay)

    def recurse(self, data={}, delay=None):
        """Have a stage invoke itself with a modified set of arguments."""
        return self.emit(stage=self.stage.name,
                         data=data,
                         delay=delay)

    def execute(self, data):
        """Execute the crawler and create a database record of having done
        so."""

        try:
            signals.operation_start.send(self)
            self.log.debug('Running: %s', self.stage.name)
            return self.stage.method(self, data)
        except Exception as exc:
            # this should clear results and tags created by this op
            # TODO: should we also use transactions on the datastore?
            session.rollback()
            self.emit_exception(exc)
        finally:
            signals.operation_end.send(self)
            shutil.rmtree(self.work_path)
            # Save the results and events created in this op
            session.commit()

    def emit_warning(self, message, type=None, details=None, *args):
        if len(args):
            message = message % args
        self.log.warning(message)
        return Event.save(self.crawler.name,
                          self.stage.name,
                          Event.LEVEL_WARNING,
                          self.run_id,
                          error_type=type,
                          error_message=message,
                          error_details=details)

    def emit_exception(self, exc):
        self.log.exception(exc)
        return Event.save(self.crawler.name,
                          self.stage.name,
                          Event.LEVEL_ERROR,
                          self.run_id,
                          error_type=exc.__class__.__name__,
                          error_message=six.text_type(exc),
                          error_details=traceback.format_exc())

    def set_tag(self, key, value):
        return Tag.save(self.crawler, key, value)

    def get_tag(self, key):
        tag = Tag.find(self.crawler, key)
        if tag is not None:
            return tag.value

    def check_tag(self, key):
        return Tag.exists(self.crawler, key)

    def skip_incremental(self, *criteria):
        """Perform an incremental check on a set of criteria.

        This can be used to execute a part of a crawler only once per an
        interval (which is specified by the ``expire`` setting). If the
        operation has already been performed (and should thus be skipped),
        this will return ``True``. If the operation needs to be executed,
        the returned value will be ``False``.
        """
        if not self.incremental:
            return False

        # this is pure convenience, and will probably backfire at some point.
        key = make_key(*criteria)
        if key is None:
            return False

        # this is used to re-run parts of a scrape after a certain interval,
        # e.g. half a year, or a year
        since = None
        if self.crawler.expire > 0:
            delta = timedelta(days=self.crawler.expire)
            since = datetime.utcnow() - delta

        if Tag.exists(self.crawler, key, since=since):
            return True
        self.set_tag(key, None)
        return False

    def store_file(self, file_path, content_hash=None):
        """Put a file into permanent storage so it can be visible to other
        stages."""
        return storage.archive_file(file_path, content_hash=content_hash)

    def store_data(self, data, encoding='utf-8'):
        """Put the given content into a file, possibly encoding it as UTF-8
        in the process."""
        path = random_filename(self.work_path)
        try:
            with open(path, 'wb') as fh:
                if isinstance(data, six.text_type):
                    data = data.encode(encoding)
                if data is not None:
                    fh.write(data)
            return self.store_file(path)
        finally:
            try:
                os.unlink(path)
            except OSError:
                pass

    @contextmanager
    def load_file(self, content_hash, file_name=None):
        file_path = storage.load_file(content_hash,
                                      file_name=file_name,
                                      temp_path=self.work_path)
        if file_path is None:
            raise StorageFileMissing(content_hash, file_name=file_name)

        try:
            with open(file_path, 'r') as fh:
                yield fh
        finally:
            storage.cleanup_file(content_hash, temp_path=self.work_path)

    def dump_state(self):
        state = deepcopy(self.state)
        state['crawler'] = self.crawler.name
        state['run_id'] = self.run_id
        return state

    @classmethod
    def from_state(cls, state, stage):
        state_crawler = state.get('crawler')
        crawler = manager.get(state_crawler)
        if crawler is None:
            raise RuntimeError("Missing crawler: [%s]" % state_crawler)
        stage = crawler.get(stage)
        if stage is None:
            raise RuntimeError('[%r] has no stage: %s' % (crawler, stage))
        return cls(crawler, stage, state)

    def __repr__(self):
        return '<Context(%r, %r)>' % (self.crawler, self.stage)


@celery.task(bind=True, max_retries=None)
def handle(task, state, stage, data):
    """Execute the operation, rate limiting allowing."""
    context = Context.from_state(state, stage)
    if context.stage.rate_limit:
        try:
            with rate_limiter(context):
                if settings.EAGER:
                    local_queue.queue_operation(context, data)
                else:
                    context.execute(data)
                return
        except RateLimitException:
            delay = max(1, 1.0/context.stage.rate_limit)
            delay = random.randint(1, int(delay))
            context.log.info("Rate limit exceeded, delaying %d sec.", delay)
            time.sleep(delay)

    if settings.EAGER:
        # If celery is running in eager mode, put the crawler in a
        # Queue. Then we get to execute them sequentially and avoid
        # recursion errors.
        local_queue.queue_operation(context, data)
    else:
        context.execute(data)



import logging
import datetime

import redis

from memorious import settings
from memorious.core import redis_pool
from memorious.signals import operation_start
from memorious.signals import operation_end
from memorious.signals import crawler_flush
from memorious.helpers.dates import parse_date

log = logging.getLogger(__name__)


def connect_redis():
    if not settings.REDIS_HOST:
        raise RuntimeError("No $MEMORIOUS_REDIS_HOST is set.")
    return redis.Redis(connection_pool=redis_pool)


def log_operation_start(context):
    conn = connect_redis()
    crawler_name = context.crawler.name
    stage_name = context.stage.name
    now = datetime.datetime.utcnow()
    conn.incr(crawler_name)
    conn.incr(crawler_name + ":" + stage_name)
    conn.incr(crawler_name + ":total_ops")
    conn.set(crawler_name + ":last_run", now)
    if not conn.sismember(crawler_name + ":runs", context.run_id):
        conn.sadd(crawler_name + ":runs", context.run_id)
        conn.set("run:" + context.run_id + ":start", now)
    conn.incr("run:" + context.run_id)
    conn.incr("run:" + context.run_id + ":total_ops")


def log_operation_end(context):
    conn = connect_redis()
    crawler_name = context.crawler.name
    conn.decr(crawler_name)
    conn.decr("run:" + context.run_id)
    if conn.get("run:" + context.run_id) == "0":
        now = datetime.datetime.utcnow()
        conn.set("run:" + context.run_id + ":end", now)


def flush_crawler(crawler):
    conn = connect_redis()
    crawler_name = crawler.name
    conn.delete(crawler_name)
    conn.delete(crawler_name + ":total_ops")
    conn.delete(crawler_name + ":last_run")
    for run_id in conn.smembers(crawler_name + ":runs"):
        conn.delete("run:" + run_id + ":start")
        conn.delete("run:" + run_id + ":end")
        conn.delete("run:" + run_id + ":total_ops")
        conn.delete("run:" + run_id)
    conn.delete(crawler_name + ":runs")
    for stage in crawler.stages:
        conn.delete(crawler_name + ":" + stage)


def cleanup_crawler(crawler):
    if not settings.REDIS_HOST:
        return None
    conn = connect_redis()
    crawler_name = crawler.name
    conn.delete(crawler_name)


def get_last_run(crawler):
    if not settings.REDIS_HOST:
        return None
    conn = connect_redis()
    last_run = conn.get(crawler.name+":last_run")
    if last_run:
        return parse_date(last_run)


def get_crawler_op_count(crawler):
    """Total operations performed for this crawler"""
    if not settings.REDIS_HOST:
        return None
    conn = connect_redis()
    total_ops = conn.get(crawler.name+":total_ops")
    if total_ops:
        return int(total_ops)


def get_stage_op_count(self):
    """Total operations performed for this stage"""
    if not settings.REDIS_HOST:
        return None
    conn = connect_redis()
    total_ops = conn.get(self.crawler.name + ":" + self.name)
    if total_ops:
        return int(total_ops)


def is_running(crawler):
    """Is the crawler currently running?"""
    if not settings.REDIS_HOST:
        return False
    conn = connect_redis()
    active_ops = conn.get(crawler.name)
    if active_ops and int(active_ops) > 0:
        return True
    return False


def get_crawler_runs(crawler):
    if not settings.REDIS_HOST:
        return []
    conn = connect_redis()
    runs = []
    for run_id in conn.smembers(crawler.name + ":runs"):
        runs.append({
            'run_id': run_id,
            'total_ops': conn.get("run:" + run_id + ":total_ops"),
            'start': parse_date(conn.get("run:" + run_id + ":start")),
            'end': parse_date(conn.get("run:" + run_id + ":end"))
        })
    return runs


def init():
    if settings.REDIS_HOST:
        log.info("redis available, configuring reporting...")
        operation_start.connect(log_operation_start)
        operation_end.connect(log_operation_end)
        crawler_flush.connect(flush_crawler)



import logging
from datetime import datetime
from sqlalchemy import Column, String, Integer, DateTime, ForeignKey

from memorious.core import session
from memorious.model.common import Base, JSON

log = logging.getLogger(__name__)


class Result(Base):
    """A resource that has been retrieved, or that failed to retrieve."""
    __tablename__ = 'result'

    id = Column(Integer, primary_key=True)
    crawler = Column(String, nullable=False, index=True)
    prev_stage = Column(String, nullable=False, index=True)
    next_stage = Column(String, nullable=False, index=True)
    data = Column(JSON, nullable=False, default={})
    timestamp = Column(DateTime, default=datetime.utcnow)

    @classmethod
    def save(cls, crawler, prev_stage, next_stage, data):
        obj = cls()
        obj.crawler = crawler.name
        obj.prev_stage = prev_stage
        obj.next_stage = next_stage
        obj.data = data
        session.add(obj)
        return obj

    @classmethod
    def delete(cls, crawler):
        pq = session.query(cls)
        pq = pq.filter(cls.crawler == crawler)
        pq.delete(synchronize_session=False)

    @classmethod
    def by_crawler_next_stage(cls, crawler, next_stage):
        q = session.query(cls)
        q = q.filter(cls.crawler == crawler)
        q = q.filter(cls.next_stage == next_stage)
        return q

    def __repr__(self):
        return '<Result(%s,%s)>' % (self.crawler, self.next_stage)



from hashlib import sha1

from memorious.util import make_key


def make_id(*parts):
    key = make_key(*parts)
    if key is not None:
        key = key.encode('utf-8')
        return sha1(key).hexdigest()
